@@ -1,4 +1,4 @@
-use std::{convert::TryInto, io::Write};
+use std::{fmt::Display, io::Write};
 
 use super::consts::{
     OPCODE_ADD, OPCODE_AND, OPCODE_BAPPEND, OPCODE_BCONS, OPCODE_BEMPTY, OPCODE_BEZ,
@@ -84,6 +84,67 @@ pub enum OpCode {
     Dup,
 }
 
+#[derive(Error, Debug, Clone)]
+pub enum ParseOpCodeError {
+    #[error("invalid opcode")]
+    InvalidOpcode,
+}
+
+impl Display for OpCode {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        match self {
+            OpCode::Noop => "noop".fmt(f),
+            OpCode::Add => "add".fmt(f),
+            OpCode::Sub => "sub".fmt(f),
+            OpCode::Mul => "mul".fmt(f),
+            OpCode::Div => "div".fmt(f),
+            OpCode::Rem => "rem".fmt(f),
+            OpCode::And => "and".fmt(f),
+            OpCode::Or => "or".fmt(f),
+            OpCode::Xor => "xor".fmt(f),
+            OpCode::Not => "not".fmt(f),
+            OpCode::Eql => "eql".fmt(f),
+            OpCode::Lt => "lt".fmt(f),
+            OpCode::Gt => "gt".fmt(f),
+            OpCode::Shl => "shl".fmt(f),
+            OpCode::Shr => "shr".fmt(f),
+            OpCode::Hash(i) => format!("hash {}", i).fmt(f),
+            OpCode::SigEOk(i) => format!("sigeok {}", i).fmt(f),
+            OpCode::Store => "store".fmt(f),
+            OpCode::Load => "load".fmt(f),
+            OpCode::StoreImm(i) => format!("storeimm {}", i).fmt(f),
+            OpCode::LoadImm(i) => format!("loadimm {}", i).fmt(f),
+            OpCode::VRef => "vref".fmt(f),
+            OpCode::VAppend => "vappend".fmt(f),
+            OpCode::VEmpty => "vempty".fmt(f),
+            OpCode::VLength => "vlength".fmt(f),
+            OpCode::VSlice => "vslice".fmt(f),
+            OpCode::VSet => "vset".fmt(f),
+            OpCode::VPush => "vpush".fmt(f),
+            OpCode::VCons => "vcons".fmt(f),
+            OpCode::BRef => "bref".fmt(f),
+            OpCode::BAppend => "bappend".fmt(f),
+            OpCode::BEmpty => "bempty".fmt(f),
+            OpCode::BLength => "blength".fmt(f),
+            OpCode::BSlice => "bslice".fmt(f),
+            OpCode::BSet => "bset".fmt(f),
+            OpCode::BPush => "bpush".fmt(f),
+            OpCode::BCons => "bcons".fmt(f),
+            OpCode::Bez(i) => format!("bez {}", i).fmt(f),
+            OpCode::Bnz(i) => format!("bnz {}", i).fmt(f),
+            OpCode::Jmp(i) => format!("jmp {}", i).fmt(f),
+            OpCode::Loop(i, j) => format!("loop {} {}", i, j).fmt(f),
+            OpCode::ItoB => "itob".fmt(f),
+            OpCode::BtoI => "btoi".fmt(f),
+            OpCode::TypeQ => "typeq".fmt(f),
+            OpCode::PushB(v) => format!("pushb {}", hex::encode(&v)).fmt(f),
+            OpCode::PushI(i) => format!("pushi {}", i).fmt(f),
+            OpCode::PushIC(i) => format!("pushic {}", i).fmt(f),
+            OpCode::Dup => "dup".fmt(f),
+        }
+    }
+}
+
 /// Opcode encoding error
 #[derive(Error, Debug)]
 pub enum EncodeError {
@@ -207,7 +268,7 @@ impl OpCode {
             OpCode::PushIC(i) => {
                 output.write_all(&[OPCODE_PUSHIC]).unwrap();
                 let bytes_repr = i.to_be_bytes();
-                let leading_zeros = i.leading_zeros() / 8;
+                let leading_zeros = bytes_repr.iter().take_while(|i| **i == 0).count();
                 output.write_all(&[32 - (leading_zeros as u8)]).unwrap();
                 output
                     .write_all(&bytes_repr[leading_zeros as usize..])
@@ -295,18 +356,19 @@ impl OpCode {
                 Ok(OpCode::PushI(U256::from_be_bytes(buf)))
             }
             OPCODE_PUSHIC => {
+                let mut buf = [0; 32];
                 let nonzero_len = read_byte(input)?;
-                let mut blit = vec![0u8; nonzero_len as usize];
+                if nonzero_len > 32 {
+                    return Err(DecodeError::InvalidVarint);
+                }
+                let mut blit = &mut buf[..nonzero_len as usize];
                 input.read_exact(&mut blit)?;
                 if blit.len() > 32 {
                     return Err(DecodeError::InvalidVarint);
                 }
                 blit.reverse();
-                let blit_len = blit.len();
-                blit.resize(32, 0);
-                // TODO TODO TODO ensure canonicalness
-                let integ = U256::from_le_bytes(blit.as_slice().try_into().unwrap());
-                if 32 - (integ.leading_zeros() / 8) != blit_len as u32 {
+                let integ = U256::from_le_bytes(buf);
+                if 32 - (integ.leading_zeros() / 8) != nonzero_len as u32 {
                     return Err(DecodeError::InvalidVarint);
                 }
                 Ok(OpCode::PushIC(integ))
